@@ -2,19 +2,19 @@
 /*
  * GData Client
  * Copyright (C) Philip Withnall 2008-2009 <philip@tecnocode.co.uk>
- * 
- * GData Client is free software: you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation, either version 3 of the License, or
- * (at your option) any later version.
+ *
+ * GData Client is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation; either
+ * version 2.1 of the License, or (at your option) any later version.
  *
  * GData Client is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with GData Client.  If not, see <http://www.gnu.org/licenses/>.
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with GData Client.  If not, see <http://www.gnu.org/licenses/>.
  */
 
 /**
@@ -40,6 +40,7 @@
 #include "gdata-service.h"
 #include "gdata-private.h"
 #include "gdata-parser.h"
+#include "atom/gdata-link.h"
 
 /* Standards reference here: http://code.google.com/apis/youtube/2.0/reference.html */
 
@@ -54,7 +55,7 @@ static void gdata_youtube_service_get_property (GObject *object, guint property_
 static void gdata_youtube_service_set_property (GObject *object, guint property_id, const GValue *value, GParamSpec *pspec);
 static gboolean parse_authentication_response (GDataService *self, guint status, const gchar *response_body, gint length, GError **error);
 static void append_query_headers (GDataService *self, SoupMessage *message);
-static void parse_error_response (GDataService *self, guint status, const gchar *reason_phrase,
+static void parse_error_response (GDataService *self, GDataServiceError error_type, guint status, const gchar *reason_phrase,
 				  const gchar *response_body, gint length, GError **error);
 
 struct _GDataYouTubeServicePrivate {
@@ -222,7 +223,8 @@ append_query_headers (GDataService *self, SoupMessage *message)
 }
 
 static void
-parse_error_response (GDataService *self, guint status, const gchar *reason_phrase, const gchar *response_body, gint length, GError **error)
+parse_error_response (GDataService *self, GDataServiceError error_type, guint status, const gchar *reason_phrase, const gchar *response_body,
+		      gint length, GError **error)
 {
 	xmlDoc *doc;
 	xmlNode *node;
@@ -234,7 +236,7 @@ parse_error_response (GDataService *self, guint status, const gchar *reason_phra
 		length = strlen (response_body);
 
 	/* Parse the XML */
-	doc = xmlReadMemory (response_body, length, "error.xml", NULL, 0);
+	doc = xmlReadMemory (response_body, length, "/dev/null", NULL, 0);
 	if (doc == NULL)
 		goto parent;
 
@@ -253,22 +255,35 @@ parse_error_response (GDataService *self, guint status, const gchar *reason_phra
 	}
 
 	/* Parse the actual errors */
-	node = node->xmlChildrenNode;
+	node = node->children;
 	while (node != NULL) {
 		xmlChar *domain = NULL, *code = NULL, *location = NULL;
-		xmlNode *child_node = node->xmlChildrenNode;
+		xmlNode *child_node = node->children;
+
+		if (node->type == XML_TEXT_NODE) {
+			/* Skip text nodes; they're all whitespace */
+			node = node->next;
+			continue;
+		}
 
 		/* Get the error data */
 		while (child_node != NULL) {
+			if (child_node->type == XML_TEXT_NODE) {
+				/* Skip text nodes; they're all whitespace */
+				child_node = child_node->next;
+				continue;
+			}
+
 			if (xmlStrcmp (child_node->name, (xmlChar*) "domain") == 0)
-				domain = xmlNodeListGetString (doc, child_node->xmlChildrenNode, TRUE);
+				domain = xmlNodeListGetString (doc, child_node->children, TRUE);
 			else if (xmlStrcmp (child_node->name, (xmlChar*) "code") == 0)
-				code = xmlNodeListGetString (doc, child_node->xmlChildrenNode, TRUE);
+				code = xmlNodeListGetString (doc, child_node->children, TRUE);
 			else if (xmlStrcmp (child_node->name, (xmlChar*) "location") == 0)
-				location = xmlNodeListGetString (doc, child_node->xmlChildrenNode, TRUE);
-			else {
-				/* Unknown element */
-				gdata_parser_error_unhandled_element ((gchar*) child_node->ns->prefix, (gchar*) child_node->name, "error", error);
+				location = xmlNodeListGetString (doc, child_node->children, TRUE);
+			else if (xmlStrcmp (child_node->name, (xmlChar*) "internalReason") != 0) {
+				/* Unknown element (ignore internalReason) */
+				g_message ("Unhandled <error/%s> element.", child_node->name);
+
 				xmlFree (domain);
 				xmlFree (code);
 				xmlFree (location);
@@ -298,11 +313,13 @@ parse_error_response (GDataService *self, guint status, const gchar *reason_phra
 						     _("You have made too many API calls recently. Please wait a few minutes and try again."));
 				} else if (xmlStrcmp (code, (xmlChar*) "too_many_entries") == 0) {
 					g_set_error (error, GDATA_YOUTUBE_SERVICE_ERROR, GDATA_YOUTUBE_SERVICE_ERROR_ENTRY_QUOTA_EXCEEDED,
-						     _("You have exceeded your entry quota with the entry \"%s\"."
-						       "Please delete some entries and try again."), location);
+						     _("You have exceeded your entry quota. Please delete some entries and try again."));
 				} else {
 					/* Protocol error */
 					g_set_error (error, GDATA_SERVICE_ERROR, GDATA_SERVICE_ERROR_PROTOCOL_ERROR,
+						     /* Translators: the first parameter is an error code, which is a coded string. The second parameter
+						      * is an error domain, which is another coded string. The third parameter is the location of the
+						      * error, which is either a URI or an XPath. */
 						     _("Unknown error code \"%s\" in domain \"%s\" received with location \"%s\"."),
 						     code, domain, location);
 				}
@@ -328,7 +345,7 @@ parse_error_response (GDataService *self, guint status, const gchar *reason_phra
 
 parent:
 	/* Chain up to the parent class */
-	GDATA_SERVICE_CLASS (gdata_youtube_service_parent_class)->parse_error_response (self, status, reason_phrase,
+	GDATA_SERVICE_CLASS (gdata_youtube_service_parent_class)->parse_error_response (self, error_type, status, reason_phrase,
 											response_body, length, error);
 	return;
 }
@@ -407,8 +424,7 @@ gdata_youtube_service_query_standard_feed (GDataYouTubeService *self, GDataYouTu
 {
 	/* TODO: Support the "time" parameter, as well as category- and region-specific feeds */
 	return gdata_service_query (GDATA_SERVICE (self), standard_feed_type_to_feed_uri (feed_type), query,
-				    (GDataEntryParserFunc) _gdata_youtube_video_new_from_xml_node, cancellable,
-				    progress_callback, progress_user_data, error);
+				    GDATA_TYPE_YOUTUBE_VIDEO, cancellable, progress_callback, progress_user_data, error);
 }
 
 /**
@@ -436,8 +452,7 @@ gdata_youtube_service_query_standard_feed_async (GDataYouTubeService *self, GDat
 						 GAsyncReadyCallback callback, gpointer user_data)
 {
 	gdata_service_query_async (GDATA_SERVICE (self), standard_feed_type_to_feed_uri (feed_type), query,
-				   (GDataEntryParserFunc) _gdata_youtube_video_new_from_xml_node, cancellable, progress_callback, progress_user_data,
-				   callback, user_data);
+				   GDATA_TYPE_YOUTUBE_VIDEO, cancellable, progress_callback, progress_user_data, callback, user_data);
 }
 
 /**
@@ -462,8 +477,7 @@ gdata_youtube_service_query_videos (GDataYouTubeService *self, GDataQuery *query
 				    GError **error)
 {
 	return gdata_service_query (GDATA_SERVICE (self), "http://gdata.youtube.com/feeds/api/videos", query,
-				    (GDataEntryParserFunc) _gdata_youtube_video_new_from_xml_node, cancellable,
-				    progress_callback, progress_user_data, error);
+				    GDATA_TYPE_YOUTUBE_VIDEO, cancellable, progress_callback, progress_user_data, error);
 }
 
 /**
@@ -490,8 +504,166 @@ gdata_youtube_service_query_videos_async (GDataYouTubeService *self, GDataQuery
 					  GAsyncReadyCallback callback, gpointer user_data)
 {
 	gdata_service_query_async (GDATA_SERVICE (self), "http://gdata.youtube.com/feeds/api/videos", query,
-				   (GDataEntryParserFunc) _gdata_youtube_video_new_from_xml_node, cancellable, progress_callback, progress_user_data,
-				   callback, user_data);
+				   GDATA_TYPE_YOUTUBE_VIDEO, cancellable, progress_callback, progress_user_data, callback, user_data);
+}
+
+/**
+ * gdata_youtube_service_query_single_video:
+ * @self: a #GDataYouTubeService
+ * @query: a #GDataQuery with the query parameters, or %NULL
+ * @video_id: the video ID of the desired video
+ * @cancellable: a #GCancellable, or %NULL
+ * @error: a #GError, or %NULL
+ *
+ * Retrieves information about the single video with the given @video_id. @video_id should be as returned by
+ * gdata_youtube_video_get_video_id() or as extracted from a YouTube URI.
+ *
+ * Parameters and errors are as for gdata_service_query().
+ *
+ * Return value: a #GDataYouTubeVideo, or %NULL; unref with g_object_unref()
+ *
+ * Since: 0.4.0
+ **/
+GDataYouTubeVideo *
+gdata_youtube_service_query_single_video (GDataYouTubeService *self, GDataQuery *query, const gchar *video_id,
+					  GCancellable *cancellable, GError **error)
+{
+	gchar *feed_uri;
+	GDataParsable *video;
+	SoupMessage *message;
+
+	g_return_val_if_fail (GDATA_IS_YOUTUBE_SERVICE (self), NULL);
+	g_return_val_if_fail (query == NULL || GDATA_IS_QUERY (query), NULL);
+	g_return_val_if_fail (video_id != NULL, NULL);
+	g_return_val_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable), NULL);
+
+	/* Query for just the specified video ID */
+	feed_uri = g_strconcat ("http://gdata.youtube.com/feeds/api/videos/", video_id, NULL);
+	message = _gdata_service_query (GDATA_SERVICE (self), feed_uri, query, cancellable, NULL, NULL, error);
+	g_free (feed_uri);
+
+	if (message == NULL)
+		return NULL;
+
+	g_assert (message->response_body->data != NULL);
+	video = gdata_parsable_new_from_xml (GDATA_TYPE_YOUTUBE_VIDEO, message->response_body->data, message->response_body->length, error);
+	g_object_unref (message);
+
+	return GDATA_YOUTUBE_VIDEO (video);
+}
+
+typedef struct {
+	GDataQuery *query;
+	gchar *video_id;
+} QuerySingleVideoAsyncData;
+
+static void
+query_single_video_async_data_free (QuerySingleVideoAsyncData *data)
+{
+	g_free (data->video_id);
+	if (data->query != NULL)
+		g_object_unref (data->query);
+	g_slice_free (QuerySingleVideoAsyncData, data);
+}
+
+static void
+query_single_video_thread (GSimpleAsyncResult *result, GDataService *service, GCancellable *cancellable)
+{
+	GDataYouTubeVideo *video;
+	GError *error = NULL;
+	QuerySingleVideoAsyncData *data = g_simple_async_result_get_op_res_gpointer (result);
+
+	/* Check to see if it's been cancelled already */
+	if (g_cancellable_set_error_if_cancelled (cancellable, &error) == TRUE) {
+		g_simple_async_result_set_from_error (result, error);
+		g_error_free (error);
+		return;
+	}
+
+	/* Execute the query and return */
+	video = gdata_youtube_service_query_single_video (GDATA_YOUTUBE_SERVICE (service), data->query, data->video_id,
+							  cancellable, &error);
+	if (video == NULL) {
+		g_simple_async_result_set_from_error (result, error);
+		g_error_free (error);
+	}
+
+	g_simple_async_result_set_op_res_gpointer (result, video, (GDestroyNotify) g_object_unref);
+}
+
+/**
+ * gdata_youtube_service_query_single_video_async:
+ * @self: a #GDataService
+ * @query: a #GDataQuery with the query parameters, or %NULL
+ * @video_id: the video ID of the desired video
+ * @cancellable: optional #GCancellable object, or %NULL
+ * @callback: a #GAsyncReadyCallback to call when authentication is finished
+ * @user_data: data to pass to the @callback function
+ *
+ * Retrieves information about the single video with the given @video_id. @video_id should be as returned by
+ * gdata_youtube_video_get_video_id() or as extracted from a YouTube URI. @self, @query and @video_id are reffed/copied when this
+ * function is called, so can safely be freed after this function returns.
+ *
+ * For more details, see gdata_youtube_service_query_single_video(), which is the synchronous version of this function.
+ *
+ * When the operation is finished, @callback will be called. You can then call gdata_youtube_service_query_single_video_finish()
+ * to get the results of the operation.
+ *
+ * Since: 0.4.0
+ **/
+void
+gdata_youtube_service_query_single_video_async (GDataYouTubeService *self, GDataQuery *query, const gchar *video_id,
+						GCancellable *cancellable, GAsyncReadyCallback callback, gpointer user_data)
+{
+	GSimpleAsyncResult *result;
+	QuerySingleVideoAsyncData *data;
+
+	g_return_if_fail (GDATA_IS_YOUTUBE_SERVICE (self));
+	g_return_if_fail (query == NULL || GDATA_IS_QUERY (query));
+	g_return_if_fail (video_id != NULL);
+	g_return_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable));
+
+	data = g_slice_new (QuerySingleVideoAsyncData);
+	data->query = (query != NULL) ? g_object_ref (query) : NULL;
+	data->video_id = g_strdup (video_id);
+
+	result = g_simple_async_result_new (G_OBJECT (self), callback, user_data, gdata_youtube_service_query_single_video_async);
+	g_simple_async_result_set_op_res_gpointer (result, data, (GDestroyNotify) query_single_video_async_data_free);
+	g_simple_async_result_run_in_thread (result, (GSimpleAsyncThreadFunc) query_single_video_thread, G_PRIORITY_DEFAULT, cancellable);
+	g_object_unref (result);
+}
+
+/**
+ * gdata_youtube_service_query_single_video_finish:
+ * @self: a #GDataYouTubeService
+ * @async_result: a #GAsyncResult
+ * @error: a #GError, or %NULL
+ *
+ * Finishes an asynchronous query operation for a single video, as started with gdata_youtube_service_query_single_video_async().
+ *
+ * Return value: a #GDataYouTubeVideo, or %NULL; unref with g_object_unref()
+ *
+ * Since: 0.4.0
+ **/
+GDataYouTubeVideo *
+gdata_youtube_service_query_single_video_finish (GDataYouTubeService *self, GAsyncResult *async_result, GError **error)
+{
+	GSimpleAsyncResult *result = G_SIMPLE_ASYNC_RESULT (async_result);
+	GDataYouTubeVideo *video;
+
+	g_return_val_if_fail (GDATA_IS_YOUTUBE_SERVICE (self), NULL);
+	g_return_val_if_fail (G_IS_ASYNC_RESULT (async_result), NULL);
+
+	g_warn_if_fail (g_simple_async_result_get_source_tag (result) == gdata_youtube_service_query_single_video_async);
+
+	if (g_simple_async_result_propagate_error (result, error) == TRUE)
+		return NULL;
+
+	video = g_simple_async_result_get_op_res_gpointer (result);
+	if (video != NULL)
+		return g_object_ref (video);
+
+	g_assert_not_reached ();
 }
 
 /**
@@ -519,18 +691,17 @@ gdata_youtube_service_query_related (GDataYouTubeService *self, GDataYouTubeVide
 	GDataLink *related_link;
 
 	/* See if the video already has a rel="http://gdata.youtube.com/schemas/2007#video.related" link */
-	related_link = gdata_entry_lookup_link (GDATA_ENTRY (video), "http://gdata.youtube.com/schemas/2007#video.related");
+	related_link = gdata_entry_look_up_link (GDATA_ENTRY (video), "http://gdata.youtube.com/schemas/2007#video.related");
 	if (related_link == NULL) {
 		/* Erroring out is probably the safest thing to do */
 		g_set_error_literal (error, GDATA_SERVICE_ERROR, GDATA_SERVICE_ERROR_PROTOCOL_ERROR,
-				     _("The video didn't have a related videos <link>."));
+				     _("The video did not have a related videos <link>."));
 		return NULL;
 	}
 
 	/* Execute the query */
-	return gdata_service_query (GDATA_SERVICE (self), related_link->href, query,
-				    (GDataEntryParserFunc) _gdata_youtube_video_new_from_xml_node, cancellable,
-				    progress_callback, progress_user_data, error);
+	return gdata_service_query (GDATA_SERVICE (self), gdata_link_get_uri (related_link), query,
+				    GDATA_TYPE_YOUTUBE_VIDEO, cancellable, progress_callback, progress_user_data, error);
 }
 
 /**
@@ -560,18 +731,17 @@ gdata_youtube_service_query_related_async (GDataYouTubeService *self, GDataYouTu
 	GDataLink *related_link;
 
 	/* See if the video already has a rel="http://gdata.youtube.com/schemas/2007#video.related" link */
-	related_link = gdata_entry_lookup_link (GDATA_ENTRY (video), "http://gdata.youtube.com/schemas/2007#video.related");
+	related_link = gdata_entry_look_up_link (GDATA_ENTRY (video), "http://gdata.youtube.com/schemas/2007#video.related");
 	if (related_link == NULL) {
 		/* Erroring out is probably the safest thing to do */
 		g_simple_async_report_error_in_idle (G_OBJECT (self), callback, user_data,
 						     GDATA_SERVICE_ERROR, GDATA_SERVICE_ERROR_PROTOCOL_ERROR,
-						     _("The video didn't have a related videos <link>."));
+						     _("The video did not have a related videos <link>."));
 		return;
 	}
 
-	gdata_service_query_async (GDATA_SERVICE (self), related_link->href, query,
-				   (GDataEntryParserFunc) _gdata_youtube_video_new_from_xml_node, cancellable, progress_callback, progress_user_data,
-				   callback, user_data);
+	gdata_service_query_async (GDATA_SERVICE (self), gdata_link_get_uri (related_link), query,
+				   GDATA_TYPE_YOUTUBE_VIDEO, cancellable, progress_callback, progress_user_data, callback, user_data);
 }
 
 /**
@@ -601,7 +771,7 @@ gdata_youtube_service_upload_video (GDataYouTubeService *self, GDataYouTubeVideo
 
 	GDataServiceClass *klass;
 	SoupMessage *message;
-	gchar *entry_xml, *upload_uri, *second_chunk_header, *upload_data, *video_contents, *i;
+	gchar *entry_xml, *second_chunk_header, *upload_data, *video_contents, *i;
 	const gchar *first_chunk_header, *footer;
 	guint status;
 	GFileInfo *video_file_info;
@@ -622,9 +792,7 @@ gdata_youtube_service_upload_video (GDataYouTubeService *self, GDataYouTubeVideo
 		return NULL;
 	}
 
-	upload_uri = g_strdup_printf ("http://uploads.gdata.youtube.com/feeds/api/users/%s/uploads", gdata_service_get_username (GDATA_SERVICE (self)));
-	message = soup_message_new (SOUP_METHOD_POST, upload_uri);
-	g_free (upload_uri);
+	message = soup_message_new (SOUP_METHOD_POST, "http://uploads.gdata.youtube.com/feeds/api/users/default/uploads");
 
 	/* Make sure subclasses set their headers */
 	klass = GDATA_SERVICE_GET_CLASS (self);
@@ -637,7 +805,7 @@ gdata_youtube_service_upload_video (GDataYouTubeService *self, GDataYouTubeVideo
 		return NULL;
 	}
 
-	entry_xml = gdata_entry_get_xml (GDATA_ENTRY (video));
+	entry_xml = gdata_parsable_get_xml (GDATA_PARSABLE (video));
 
 	/* Check for cancellation */
 	if (g_cancellable_set_error_if_cancelled (cancellable, error) == TRUE) {
@@ -702,7 +870,11 @@ gdata_youtube_service_upload_video (GDataYouTubeService *self, GDataYouTubeVideo
 	soup_message_set_request (message, "multipart/related; boundary=" BOUNDARY_STRING, SOUP_MEMORY_TAKE, upload_data, content_length);
 
 	/* Send the message */
-	status = soup_session_send_message (gdata_service_get_session (GDATA_SERVICE (self)), message);
+	status = _gdata_service_send_message (GDATA_SERVICE (self), message, error);
+	if (status == SOUP_STATUS_NONE) {
+		g_object_unref (message);
+		return NULL;
+	}
 
 	/* Check for cancellation */
 	if (g_cancellable_set_error_if_cancelled (cancellable, error) == TRUE) {
@@ -712,7 +884,7 @@ gdata_youtube_service_upload_video (GDataYouTubeService *self, GDataYouTubeVideo
 
 	if (status != 201) {
 		/* Error */
-		parse_error_response (GDATA_SERVICE (self), status, message->reason_phrase,
+		parse_error_response (GDATA_SERVICE (self), GDATA_SERVICE_ERROR_WITH_INSERTION, status, message->reason_phrase,
 				      message->response_body->data, message->response_body->length, error);
 		g_object_unref (message);
 		return NULL;
@@ -720,7 +892,8 @@ gdata_youtube_service_upload_video (GDataYouTubeService *self, GDataYouTubeVideo
 
 	g_assert (message->response_body->data != NULL);
 
-	return gdata_youtube_video_new_from_xml (message->response_body->data, (gint) message->response_body->length, error);
+	return GDATA_YOUTUBE_VIDEO (gdata_parsable_new_from_xml (GDATA_TYPE_YOUTUBE_VIDEO, message->response_body->data,
+								 (gint) message->response_body->length, error));
 }
 
 /**
