@@ -101,41 +101,51 @@ test_authentication_async (void)
 static void
 test_upload_simple (GDataService *service)
 {
-	GDataCategory *category;
 	GDataPicasaWebFile *photo;
 	GDataPicasaWebFile *photo_new;
 	GFile *photo_file;
 	gchar *xml;
 	GError *error = NULL;
+	GTimeVal timeval;
+	gchar *time_str;
+	gchar *summary;
+	gchar *expected_xml;
+
+	g_get_current_time (&timeval);
+	time_str = g_time_val_to_iso8601 (&timeval);
+	summary = g_strdup_printf ("Photo Summary (%s)", time_str);
+	g_free (time_str);
+
+	expected_xml = g_strdup_printf ("<entry "
+						"xmlns='http://www.w3.org/2005/Atom' "
+						"xmlns:gphoto='http://schemas.google.com/photos/2007' "
+						"xmlns:media='http://search.yahoo.com/mrss/' "
+						"xmlns:gd='http://schemas.google.com/g/2005' "
+						"xmlns:exif='http://schemas.google.com/photos/exif/2007' "
+						"xmlns:app='http://www.w3.org/2007/app' "
+						"xmlns:georss='http://www.georss.org/georss' "
+						"xmlns:gml='http://www.opengis.net/gml'>"
+						"<title type='text'>Photo Entry Title</title>"
+						"<summary type='text'>%s</summary>"
+						"<gphoto:position>0</gphoto:position>"
+						"<gphoto:commentingEnabled>true</gphoto:commentingEnabled>"
+						"<media:group>"
+							"<media:title type='plain'>Photo Entry Title</media:title>"
+							"<media:description type='plain'>%s</media:description>"
+						"</media:group>"
+					"</entry>", summary, summary);
 
 	photo = gdata_picasaweb_file_new (NULL);
 
 	gdata_entry_set_title (GDATA_ENTRY (photo), "Photo Entry Title");
-	gdata_picasaweb_file_set_caption (photo, "Photo Summary");
-	/* TODO: Have it add this category automatically? Same for GDataCalendarEvent */
-	category = gdata_category_new ("http://schemas.google.com/photos/2007#photo", "http://schemas.google.com/g/2005#kind", NULL);
-	gdata_entry_add_category (GDATA_ENTRY (photo), category);
+	gdata_picasaweb_file_set_caption (photo, summary);
 
 	/* Check the XML */
 	xml = gdata_parsable_get_xml (GDATA_PARSABLE (photo));
-	g_assert_cmpstr (xml, ==,
-			 "<entry "
-				"xmlns='http://www.w3.org/2005/Atom' "
-				"xmlns:gphoto='http://schemas.google.com/photos/2007' "
-				"xmlns:media='http://video.search.yahoo.com/mrss' "
-				"xmlns:gd='http://schemas.google.com/g/2005' "
-				"xmlns:app='http://www.w3.org/2007/app'>"
-				"<title type='text'>Photo Entry Title</title>"
-				"<summary type='text'>Photo Summary</summary>"
-				"<category term='http://schemas.google.com/photos/2007#photo' scheme='http://schemas.google.com/g/2005#kind'/>"
-				"<gphoto:position>0</gphoto:position>"
-				"<gphoto:commentingEnabled>true</gphoto:commentingEnabled>"
-				"<media:group>"
-					"<media:title type='plain'>Photo Entry Title</media:title>"
-					"<media:description type='plain'>Photo Summary</media:description>"
-				"</media:group>"
-			 "</entry>");
-	g_free(xml);
+	g_assert_cmpstr (xml, ==, expected_xml);
+	g_free (xml);
+
+	gdata_picasaweb_file_set_coordinates (photo, 17.127, -110.35);
 
 	/* File is public domain: http://en.wikipedia.org/wiki/File:German_garden_gnome_cropped.jpg */
 	photo_file = g_file_new_for_path (TEST_FILE_DIR "photo.jpg");
@@ -149,12 +159,13 @@ test_upload_simple (GDataService *service)
 
 	/* TODO: check entries and feed properties */
 
+	g_free (summary);
+	g_free (expected_xml);
 	g_object_unref (photo);
 	g_object_unref (photo_new);
 	g_object_unref (photo_file);
 }
 
-
 static void
 test_photo (GDataService *service)
 {
@@ -172,6 +183,11 @@ test_photo (GDataService *service)
 	GDataMediaThumbnail *thumbnail;
 	GTimeVal _time;
 	gchar *str;
+	gchar *timestamp;
+	gdouble latitude;
+	gdouble longitude;
+	gdouble original_latitude;
+	gdouble original_longitude;
 
 	album_feed = gdata_picasaweb_service_query_all_albums (GDATA_PICASAWEB_SERVICE (service), NULL, NULL, NULL, NULL, NULL, &error);
 	g_assert_no_error (error);
@@ -210,9 +226,8 @@ test_photo (GDataService *service)
 	/* TODO: file wasn't uploaded with checksum assigned; g_assert_cmpstr (gdata_picasaweb_file_get_checksum (photo), ==, ??); */
 
 	gdata_picasaweb_file_get_timestamp (photo, &_time);
-	str = g_time_val_to_iso8601 (&_time);
-	g_assert_cmpstr (str, ==, "2008-12-06T18:32:10Z");
-	g_free (str);
+	timestamp = g_time_val_to_iso8601 (&_time);
+	g_assert_cmpstr (timestamp, ==, "2008-12-06T18:32:10Z");
 
 	g_assert_cmpstr (gdata_picasaweb_file_get_video_status (photo), ==, NULL);
 	/* todo: not a good test of video status; want to upload a video for it */
@@ -246,6 +261,39 @@ test_photo (GDataService *service)
 	g_assert_cmpuint (gdata_media_thumbnail_get_width (thumbnail), ==, 288);
 	g_assert_cmpuint (gdata_media_thumbnail_get_height (thumbnail), ==, 216);
 	/* TODO consider testing time, gint64 */
+
+	/* Check EXIF values */
+	g_assert_cmpfloat (gdata_picasaweb_file_get_distance (photo), ==, 0);
+	g_assert_cmpfloat (gdata_picasaweb_file_get_exposure (photo), ==, 0.016666668);
+	g_assert_cmpint (gdata_picasaweb_file_get_flash (photo), ==, TRUE);
+	g_assert_cmpfloat (gdata_picasaweb_file_get_focal_length (photo), ==, 6.3);
+	g_assert_cmpfloat (gdata_picasaweb_file_get_fstop (photo), ==, 2.8);
+	g_assert_cmpstr (gdata_picasaweb_file_get_image_unique_id (photo), ==, "1c179e0ac4f6741c8c1cdda3516e69e5");
+	g_assert_cmpint (gdata_picasaweb_file_get_iso (photo), ==, 80);
+	g_assert_cmpstr (gdata_picasaweb_file_get_make (photo), ==, "EASTMAN KODAK COMPANY");
+	g_assert_cmpstr (gdata_picasaweb_file_get_model (photo), ==, "KODAK Z740 ZOOM DIGITAL CAMERA");
+
+	/* Check GeoRSS coordinates */
+	gdata_picasaweb_file_get_coordinates (photo, &original_latitude, &original_longitude);
+	g_assert_cmpfloat (original_latitude, ==, 45.4341173);
+	g_assert_cmpfloat (original_longitude, ==, 12.1289062);
+
+	gdata_picasaweb_file_get_coordinates (photo, NULL, &longitude);
+	g_assert_cmpfloat (longitude, ==, 12.1289062);
+	gdata_picasaweb_file_get_coordinates (photo, &latitude, NULL);
+	g_assert_cmpfloat (latitude, ==, 45.4341173);
+	gdata_picasaweb_file_get_coordinates (photo, NULL, NULL);
+
+	gdata_picasaweb_file_set_coordinates (photo, original_longitude, original_latitude);
+	gdata_picasaweb_file_get_coordinates (photo, &latitude, &longitude);
+	g_assert_cmpfloat (latitude, ==, original_longitude);
+	g_assert_cmpfloat (longitude, ==, original_latitude);
+	gdata_picasaweb_file_set_coordinates (photo, original_latitude, original_longitude);
+	gdata_picasaweb_file_get_coordinates (photo, &latitude, &longitude);
+	g_assert_cmpfloat (latitude, ==, 45.4341173);
+	g_assert_cmpfloat (longitude, ==, 12.1289062);
+
+	g_free (timestamp);
 }
 
 static void
@@ -349,6 +397,11 @@ test_album (GDataService *service)
 	GList *albums;
 	GTimeVal _time;
 	gchar *str;
+	gdouble latitude;
+	gdouble longitude;
+	gdouble original_latitude;
+	gdouble original_longitude;
+	gchar *original_rights;
 
 	album_feed = gdata_picasaweb_service_query_all_albums (GDATA_PICASAWEB_SERVICE (service), NULL, NULL, NULL, NULL, NULL, &error);
 	g_assert_no_error (error);
@@ -369,7 +422,7 @@ test_album (GDataService *service)
 
 	g_assert_cmpstr (gdata_picasaweb_album_get_description (album), ==, "This is the test description.  This album should be in Venice.");
 	g_assert_cmpint (gdata_picasaweb_album_get_visibility (album), ==, GDATA_PICASAWEB_PUBLIC);
-	/* Google doesn't seem to be returning this one any more */
+	/* TODO: Google doesn't seem to be returning this one any more; investigate */
 	/*g_assert_cmpstr (gdata_picasaweb_album_get_name (album), ==, "TestAlbum1VenicePublic");*/
 	g_assert_cmpstr (gdata_picasaweb_album_get_location (album), ==, "Venice");
 
@@ -381,6 +434,51 @@ test_album (GDataService *service)
 	g_assert_cmpuint (gdata_picasaweb_album_get_num_photos (album), ==, 1);
 	g_assert_cmpuint (gdata_picasaweb_album_get_num_photos_remaining (album), ==, 499);
 	g_assert_cmpuint (gdata_picasaweb_album_get_bytes_used (album), ==, 1124730);
+
+	gdata_picasaweb_album_get_coordinates (album, &latitude, &longitude);
+	g_assert_cmpfloat (latitude, ==, 45.434336);
+	gdata_picasaweb_album_get_coordinates (album, &original_latitude, &original_longitude);
+	g_assert_cmpfloat (original_latitude, ==, 45.434336);
+	g_assert_cmpfloat (original_longitude, ==, 12.338784);
+
+	gdata_picasaweb_album_get_coordinates (album, NULL, &longitude);
+	g_assert_cmpfloat (longitude, ==, 12.338784);
+	gdata_picasaweb_album_get_coordinates (album, &latitude, NULL);
+	g_assert_cmpfloat (latitude, ==, 45.434336);
+	gdata_picasaweb_album_get_coordinates (album, NULL, NULL);
+
+	gdata_picasaweb_album_set_coordinates (album, original_longitude, original_latitude);
+	gdata_picasaweb_album_get_coordinates (album, &latitude, &longitude);
+	g_assert_cmpfloat (latitude, ==, original_longitude);
+	g_assert_cmpfloat (longitude, ==, original_latitude);
+	gdata_picasaweb_album_set_coordinates (album, original_latitude, original_longitude);
+	gdata_picasaweb_album_get_coordinates (album, &original_latitude, &original_longitude);
+	g_assert_cmpfloat (original_latitude, ==, 45.434336);
+	g_assert_cmpfloat (original_longitude, ==, 12.338784);
+
+	/* Test visibility and its synchronisation with its GDataEntry's rights */
+	original_rights = g_strdup (gdata_entry_get_rights (GDATA_ENTRY (album)));
+
+	gdata_entry_set_rights (GDATA_ENTRY (album), "private");
+	g_assert_cmpstr (gdata_entry_get_rights (GDATA_ENTRY (album)), ==, "private");
+	g_assert_cmpint (gdata_picasaweb_album_get_visibility (album), ==, GDATA_PICASAWEB_PRIVATE);
+
+	gdata_entry_set_rights (GDATA_ENTRY (album), "public");
+	g_assert_cmpstr (gdata_entry_get_rights (GDATA_ENTRY (album)), ==, "public");
+	g_assert_cmpint (gdata_picasaweb_album_get_visibility (album), ==, GDATA_PICASAWEB_PUBLIC);
+
+	gdata_picasaweb_album_set_visibility (album, GDATA_PICASAWEB_PRIVATE);
+	g_assert_cmpstr (gdata_entry_get_rights (GDATA_ENTRY (album)), ==, "private");
+	g_assert_cmpint (gdata_picasaweb_album_get_visibility (album), ==, GDATA_PICASAWEB_PRIVATE);
+
+	gdata_picasaweb_album_set_visibility (album, GDATA_PICASAWEB_PUBLIC);
+	g_assert_cmpstr (gdata_entry_get_rights (GDATA_ENTRY (album)), ==, "public");
+	g_assert_cmpint (gdata_picasaweb_album_get_visibility (album), ==, GDATA_PICASAWEB_PUBLIC);
+
+	gdata_entry_set_rights (GDATA_ENTRY (album), original_rights);
+	g_free (original_rights);
+
+	g_object_unref (album_feed);
 }
 
 static void
@@ -411,6 +509,7 @@ test_album_feed_entry (GDataService *service)
 	g_assert_cmpstr (gdata_entry_get_title (entry), ==, "Test Album 1 - Venice - Public");
 	g_assert_cmpstr (gdata_entry_get_id (entry), ==, "5328889949261497249");
 	g_assert_cmpstr (gdata_entry_get_etag (entry), !=, NULL);
+	g_assert_cmpstr (gdata_entry_get_rights (entry), ==, "public");
 
 	gdata_entry_get_updated (entry, &_time);
 	str = g_time_val_to_iso8601 (&_time);
@@ -422,9 +521,6 @@ test_album_feed_entry (GDataService *service)
 	g_assert_cmpstr (str, ==, "2009-04-26T07:00:00Z");
 	g_free (str);
 
-	/* g_assert_cmpstr (gdata_entry_get_content (entry), !=, NULL); */
-	/* TODO */
-	printf("** WARNING:%s:%d: gdata_entry_get_content(entry) returns null; valid?\n", __FILE__, __LINE__);
 	xml = gdata_parsable_get_xml (GDATA_PARSABLE (entry));
 	g_assert_cmpstr (xml, !=, NULL);
 	g_assert_cmpuint (strlen (xml), >, 0);
@@ -458,21 +554,27 @@ test_album_feed (GDataService *service)
 static void
 test_query_all_albums (GDataService *service)
 {
-	GDataFeed *album_feed;
-	GDataFeed *photo_feed;
+	GDataFeed *album_feed, *photo_feed;
+	GDataQuery *query;
 	GError *error = NULL;
 	GList *albums;
 	GDataEntry *entry;
 	GDataPicasaWebAlbum *album;
 
-	/* TODO: find out whether I need to free this; probably */
+	/* Test a query with a "q" parameter; it should fail */
+	query = GDATA_PICASAWEB_QUERY (gdata_picasaweb_query_new ("foobar"));
+	album_feed = gdata_picasaweb_service_query_all_albums (GDATA_PICASAWEB_SERVICE (service), query, NULL, NULL, NULL, NULL, &error);
+	g_assert_error (error, GDATA_SERVICE_ERROR, GDATA_SERVICE_ERROR_BAD_QUERY_PARAMETER);
+	g_assert (album_feed == NULL);
+	g_clear_error (&error);
+
+	/* Now try a proper query */
 	album_feed = gdata_picasaweb_service_query_all_albums (GDATA_PICASAWEB_SERVICE (service), NULL, NULL, NULL, NULL, NULL, &error);
 	g_assert_no_error (error);
 	g_assert (GDATA_IS_FEED (album_feed));
 	g_clear_error (&error);
 
 	albums = gdata_feed_get_entries (album_feed);
-	/* g_object_unref (feed); TODO find out why this complains about not being an object */
 	entry = GDATA_ENTRY (g_list_nth_data (albums, TEST_ALBUM_INDEX));
 	album = GDATA_PICASAWEB_ALBUM (entry);
 
@@ -482,17 +584,15 @@ test_query_all_albums (GDataService *service)
 	g_assert (GDATA_IS_FEED (photo_feed));
 	g_clear_error (&error);
 
-	/*g_object_unref(photo_feed);
-	g_object_unref(album_feed);*/
-
-	g_list_free(albums);
+	g_object_unref (photo_feed);
+	g_object_unref (album_feed);
 }
 
 static void
 test_query_all_albums_async_cb (GDataService *service, GAsyncResult *async_result, GMainLoop *main_loop)
 {
 	GDataFeed *feed;
-	GError *error;
+	GError *error = NULL;
 
 	feed = gdata_service_query_finish (service, async_result, &error);
 	g_assert_no_error (error);
@@ -518,6 +618,7 @@ test_query_all_albums_async (GDataService *service)
 }
 
 /* TODO: test private, public albums, test uploading */
+/* TODO: add queries to update albums, files on the server; test those */
 
 int
 main (int argc, char *argv[])
@@ -536,6 +637,7 @@ main (int argc, char *argv[])
 	g_test_add_func ("/picasaweb/authentication", test_authentication);
 	if (g_test_thorough () == TRUE)
 		g_test_add_func ("/picasaweb/authentication_async", test_authentication_async);
+	g_test_add_data_func ("/picasaweb/upload/photo", service, test_upload_simple);
 	g_test_add_data_func ("/picasaweb/query/all_albums", service, test_query_all_albums);
 	if (g_test_thorough () == TRUE)
 		g_test_add_data_func ("/picasaweb/query/all_albums_async", service, test_query_all_albums_async);
@@ -545,7 +647,6 @@ main (int argc, char *argv[])
 	g_test_add_data_func ("/picasaweb/query/photo_feed", service, test_photo_feed);
 	g_test_add_data_func ("/picasaweb/query/photo_feed_entry", service, test_photo_feed_entry);
 	g_test_add_data_func ("/picasaweb/query/photo", service, test_photo);
-	g_test_add_data_func ("/picasaweb/upload/photo", service, test_upload_simple);
 
 	retval = g_test_run ();
 	g_object_unref (service);
